/*
 * getch: get a character.
 *
 * needs ltoa2()
 *
 * 92/08/22 john h. dubois iii (john@armory.com)
 * 93/02/08 Added -p option
 * 96/06/25 Added -hEsn and radix options.  Let stty options be given.
 */
#ifdef _M_SYSV
#define M_SYSV
#endif

/* Header inclusion */
#ifdef M_SYSV
#include <termio.h>
#include <unistd.h>
#else
#include <sgtty.h>
#endif
#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>

/* Global vars */
#ifdef M_SYSV
struct termio OrigMode;
#else
struct sgttyb OrigMode;
#endif
char ident[] = "@(#) getch 1.1 96/06/25";
int IsTTY;
int Debug = 0;

/* Macro declarations, etc. */
#ifdef M_SYSV
/* externalTTYmode: set TTY mode to that in effect when the program was
   started, or the last time we returned from being suspended */
#define externalTTYmode(void) ioctl(0,TCSETA,&OrigMode)
#define voidt void
#else
#define voidt int
#define void
#define externalTTYmode() ioctl(0,TIOCSETP,&OrigMode)
#endif

/* Function prototypes/declarations */
#ifdef M_UNIX
void Suspend();
#endif
char *ltoa2(long inval, char *string, int radix, int length);
int internalTTYmode(int getNewMode, int EchoChars, int KeySig, char *argv[]);
int dostty(char **argv);
void errExit(int e);
int spawn(char *cmd, char *argv[]);

void CleanExit(int i)
{
    errExit(0);
}

void errExit(int e)
{
    if (Debug)
	WriteStrings(2,"Restoring TTY to external mode\n",NULL);
    if (IsTTY)
	externalTTYmode(void);
    exit(e);
}

#ifdef M_SYSV
#ifdef M_UNIX
void SetSigs(void)
{
    struct sigaction act;

    act.sa_flags = 0;
    sigemptyset(&act.sa_mask);
    act.sa_handler = Suspend;
    sigaction(SIGTSTP,&act,(void *) 0);
    act.sa_handler = CleanExit;
    sigaction(SIGINT,&act,(void *) 0);
    sigaction(SIGQUIT,&act,(void *) 0);
    sigaction(SIGTERM,&act,(void *) 0);
    sigaction(SIGPIPE,&act,(void *) 0);
}
#else
void SetSigs()
{
    signal(SIGINT,CleanExit);
    signal(SIGQUIT,CleanExit);
    signal(SIGTERM,CleanExit);
    signal(SIGPIPE,CleanExit);
}
#endif
#else
void SetSigs(void)
{
    signal(SIGTSTP,Suspend);
    signal(SIGINT,CleanExit);
    signal(SIGQUIT,CleanExit);
    signal(SIGTERM,CleanExit);
    signal(SIGPIPE,CleanExit);
}
#endif

int spawn(char *cmd, char *argv[])
{
    int waitinfo;

    switch (fork()) {
    case -1:
	return(-1);
	break;
    case 0:
	execv(cmd,argv);
	fputs("Could not run ",stderr);
	fflush(stderr);
	perror(cmd);
	exit(1);
    default:
	return((wait(&waitinfo) < 0) ? -2 : WEXITSTATUS(waitinfo));
    }
}

int dostty(char **argv)
{
    int ret;
    argv--;
    argv[0] = "stty";
    switch (ret = spawn("/bin/stty",argv)) {
    case 0:
	return 0;
    case -1:
	perror("Could not run stty: fork failed");
	return -1;
    case -2:
	perror("Error running stty");
	return -1;
    default:
	fprintf(stderr,"stty failed: exit status %d.\n",ret);
	return 1;
    }
}

/*
*   internalTTYmode():
*   Save original (external) mode in OrigMode.
*   Set mode to that used internally.
*   This is used to initially set the TTY mode to what we want,
*   and also reset to what we want after having been suspended.
*/
#ifdef M_SYSV
int internalTTYmode(int getNewMode, int EchoChars, int KeySig, char *argv[])
{

    static struct termio NewMode;

    /* Get external mode at start and every time we return from suspend */
    if (ioctl(0,TCGETA,&OrigMode) == -1) {
	perror("ioctl to get TTY mode failed");
	return 0;
    }

    SetSigs();

    /* Only build NewMode once since it may require running stty */
    if (getNewMode) {
	NewMode = OrigMode;
	NewMode.c_lflag &= ~(ICANON | (EchoChars ? 0 : ECHO) |
	(KeySig ? 0 : ISIG));
	NewMode.c_cc[VMIN] = 1;
	NewMode.c_cc[VTIME] = 0;
    }
    if (Debug)
	WriteStrings(2,"Setting TTY to internal mode\n",NULL);
    if (ioctl(0,TCSETA,&NewMode) == -1) {
	perror("ioctl to set internal TTY mode failed");
	return 0;
    }
    if (getNewMode && argv) {
	if (dostty(argv))
	    return 0;
	if (ioctl(0,TCGETA,&NewMode) == -1) {
	    perror("ioctl to get stty-modified TTY mode failed");
	    return 0;
	}
    }
    return 1;
}
#else
int internalTTYmode(int getNewMode, int EchoChars, int KeySig, char *argv[])
{
    static struct sgttyb NewMode;
    static char ErrMsg[] = "ioctl failed.\n";

    /* Get external mode at start and every time we return from suspend */
    ioctl(0,TIOCGETP,&OrigMode);

    SetSigs();

    /* Only build NewMode once since it may require running stty */
    if (getNewMode) {
	NewMode = OrigMode;
	NewMode.sg_flags |= CBREAK;
	if (!EchoChars)
	    NewMode.sg_flags &= ~ECHO;
	if (!KeySig)
	    NewMode.sg_flags &= ~ISIG;	/* untested */
    }
    if (ioctl(0,TIOCSETP,&NewMode) == -1) {
	write(1,ErrMsg,strlen(ErrMsg));
	return 0;
    }
    if (getNewMode) {
	if (argv && dostty(argv))
	    return 0;
	ioctl(0,TIOCGETP,&NewMode);
    }
    return 1;
}
#endif

#ifdef M_SYSV
#ifdef M_UNIX
void Suspend()
{
    externalTTYmode(void);
    kill(getpid(),SIGSTOP);
    if (!internalTTYmode(0,0,0,(char **)0))
	errExit(1);
}
#endif
#else
void Suspend()
{
    externalTTYmode(void);
    kill(getpid(),SIGSTOP);
    if (!internalTTYmode(0,0,0,(char **)0))
	errExit(1);
}
#endif

getEOFchar()
{
#ifdef M_SYSV
	return OrigMode.c_cc[VEOF];
#else
	struct tchars stchars;
	ioctl(0,TIOCGETC,&stchars);
	return stchars.t_eofc;
#endif
}

void main(argc,argv)
int argc;
char *argv[];
{
    static unsigned char buf[] = " \n";
    extern int errno;
    int Num;
    int Continuous = 0;
    char **p;
    char EOFChar;
    int Count = 0;
    int Radix = 0;
    char ltoaBuf[34];
    int Length = 0;
    int doEOF = 1;
    static char ErrMsg[] = "Invalid option.  Use -h for help.\n";
    extern char *optarg;        /* For optargs */
    extern int optind;          /* For optargs */
    char c;
    int EchoChars = 0;
    int KeySig = 1;

    static char *UsageMsg[] = {
"getch: read a character in raw mode and write it and a newline to stdout.\n",
"If the input is a TTY, it is put in non-canonical (character-by-character)\n",
"mode to read a single character, and characters are not echoed as they are\n",
"typed.  getch will trap keyboard generated signals as well as the\n",
"terminate signal, and restore the TTY to the mode in effect before it was\n",
"run.  Other signals may cause getch to leave the TTY in non-canonical mode.\n",
"If the input is a TTY and the EOF character is read, getch will exit\n",
"without writing anything.\n",
"Usage: getch [-bcdeEhHops] [-n<num>] [-r<radix>] [-- stty-option ...]\n",
"If any stty-options are given, they are passed to stty for processing.\n",
"Whatever modes they set are unset when getch exits.  If the first\n",
"stty-option starts with a '-', use -- to terminate the argument list\n",
"before the first stty-option is given.  Example:\n",
"getch -sE -n10 -- -brkint\n",
"Options:\n"\
"-e: Do not disable echoing of characters typed to TTY.\n",
"-c: Get and write characters until stopped by signal or end of input.  In\n",
"    this mode, getch will terminate if sent a signal or if the input ends.\n",
"    If the input is a TTY, getch will also terminate if the EOF character\n",
"    is read.  Also, getch can be stopped by a broken pipe signal, which\n",
"    will be delivered if getch is writing to a pipe and the output of the\n",
"    pipe is closed.\n",
"-E: Treat the EOF character like any other character.\n",
"-h: Print this help.\n",
"-p: Print a line giving the process id.  This can be used to kill\n",
"    getch without it reading & writing another character.\n",
"-b: Print characters as 8-digit binary values.\n",
"-o: Print characters as 3-digit octal values.\n",
"-d: Print characters as 3-digit decimal values.\n",
"-H: Print characters as 2-digit hexadecimal values.\n",
"-r<radix>: Print characters as numbers with radix <radix>.\n",
"-s: Disable keyboard signals.  If this is given with -E, getch is liable\n",
"    to be difficult to quit.\n",
"-n<num>: Like -c, but also quit after writing <num> characters.\n",
(char *) 0 };
    
    while ((c = getopt(argc,argv,"bcdeEhHopsn:r:x:")) != -1)
	switch (c) {
	case 'e':
	    EchoChars = 1;
	    break;
	case 'E':
	    doEOF = 0;
	    break;
	case 'c':
	    Continuous = 1;
	    break;
	case '-':
	    break;
	case 'b':
	    Radix = 2;
	    Length = 8;
	    break;
	case 'o':
	    Radix = 8;
	    Length = 3;
	    break;
	case 'd':
	    Radix = 10;
	    Length = 3;
	    break;
	case 'H':
	    Radix = 16;
	    Length = 2;
	    break;
	case 'r':
	    Radix = atoi(optarg);
	    if (!(2 <= Radix && Radix <= 36)) {
		static char rErr[] = "Radix must be between 2 and 36.\n";
		write(2,rErr,strlen(rErr));
		exit(1);
	    }
	    break;
	case 'x':
	    Debug = atoi(optarg);
	    fprintf(stderr,"Debugging set to level %d\n",Debug);
	    break;
	case 's':
	    KeySig = 0;
	    break;
	case 'p':
	    ltoa2((long) getpid(),ltoaBuf,10,0);
	    write(1,ltoaBuf,strlen(ltoaBuf));
	    write(1,"\n",1);
	    break;
	case 'n':
	    Continuous = 1;
	    Count = atoi(optarg);
	    if (!Count) {
		static char nErr[] =
		"Value given with -n must be a positive number.\n";
		write(2,nErr,strlen(nErr));
		exit(1);
	    }
	    break;
	case 'h':
	    for (p = UsageMsg; *p; p++)
		write(1,*p,strlen(*p));
	    exit(0);
	default:
	    write(2,ErrMsg,strlen(ErrMsg));
	    exit(1);
	    break;
	}
    argc -= optind;
    IsTTY = isatty(0);
    if (IsTTY) {
	if (!internalTTYmode(1,EchoChars,KeySig,argc ? argv+optind:(char **)0))
	    errExit(1);
	if (doEOF)
	    EOFChar = getEOFchar();
    }
    else
	doEOF = 0;
    do {
	while ((Num = read(0,buf,1)) == -1)
	    if (errno != EINTR) 
		CleanExit(0);
	if (!Num || doEOF && *buf == EOFChar)
	    CleanExit(0);
	if (Radix) {
	    ltoa2((long) *(buf + 1 - Num), ltoaBuf, Radix, Length);
	    write(1,ltoaBuf,strlen(ltoaBuf));
	    write(1,"\n",1);
	}
	else
	    write(1,buf + 1 - Num,2);
     
    } while (Continuous && (!Count || --Count));
    CleanExit(0);
}
